# ------------------------------ LOAD LIBRARIES -------------------------------
import time
start_time = time.time()

from netgen.geom2d import unit_square
from ngsolve import *
from ngsolve.krylovspace import GMRes, PreconditionedRichardson
from math import log, pi

SetNumThreads(4)
# -------------------------------- PARAMETERS ---------------------------------
nu = 1e-4

h0 = 0.3
lvls = range(0,6)
k = 4
pReg = 1e-8

direct_solver = False
precon = "bddc"
inverse = "umfpack"
condense = True

# ----------------------------------- DATA ------------------------------------
u_ex = CoefficientFunction((sin(2*pi*x)*sin(2*pi*y),cos(2*pi*x)*cos(2*pi*y)))
p_ex = (cos(4*pi*x)-cos(4*pi*y))/4

grad_u1_ex = CoefficientFunction((	2*pi*cos(2*pi*x)*sin(2*pi*y),
									2*pi*sin(2*pi*x)*cos(2*pi*y)))
grad_u2_ex = CoefficientFunction((	-2*pi*sin(2*pi*x)*cos(2*pi*y),
									-2*pi*cos(2*pi*x)*sin(2*pi*y)))
grad_u_ex = CoefficientFunction((grad_u1_ex,grad_u2_ex))


f = 8*nu*pi*pi*u_ex
f += pi*CoefficientFunction((-sin(4*pi*x),sin(4*pi*y)))


# ----------------------------------- MESH ------------------------------------
mesh = Mesh(unit_square.GenerateMesh(maxh=h0))

# --------------------------- FINITE ELEMENT SPACE ----------------------------
V = VectorH1(mesh, order=k, dirichlet="top|left|right|bottom")
Q = H1(mesh,order=k-1)
X = FESpace([V,Q])

gfu = GridFunction(X)
vel,pre = gfu.components

# ----------------------------- (BI)LINEAR FORMS ------------------------------
(u,p),(v,q) = X.TnT()

stokes = nu*InnerProduct(grad(u),grad(v)) - p*div(v) - q*div(u) - pReg*p*q
rhs = InnerProduct(f,v)

# -------------------------------- INTEGRATORS --------------------------------
a = BilinearForm(X,condense=condense)
a += SymbolicBFI(form = stokes)

if not direct_solver:
	c = Preconditioner( a, precon, inverse=inverse)

f = LinearForm(X)
f += SymbolicLFI(form = rhs)

# ----------------------------- ERROR COMPUTATION -----------------------------
Errors = {}
Errors["L2v"] = []
Errors["H1v"] = []
Errors["L2p"] = []

def CompError():
	l2v_err = sqrt(Integrate(InnerProduct(vel-u_ex,vel-u_ex),mesh))
	h1v_err = sqrt(Integrate(InnerProduct(
								grad(vel)-grad_u_ex,grad(vel)-grad_u_ex),mesh))
	l2p_err = sqrt(Integrate(InnerProduct(pre-p_ex,pre-p_ex),mesh))

	Errors["L2v"].append(l2v_err)
	Errors["H1v"].append(h1v_err)
	Errors["L2p"].append(l2p_err)

# ---------------------------------- VISUALS ----------------------------------
Draw(vel,mesh,"u")
Draw(pre,mesh,"p")

# ------------------------------- SOLVE PROBLEM -------------------------------
for l in lvls:
	with TaskManager():
		if l > 0 : mesh.Refine()

		X.Update()
		a.Assemble()
		
		f.Assemble()
		gfu.Update()

		vel.Set(u_ex, definedon=mesh.Boundaries("top|left|bottom|right"))
		f.vec.data += -a.mat*gfu.vec
		if condense:
			f.vec.data += a.harmonic_extension_trans * f.vec

		freedofs = X.FreeDofs(coupling=condense)

		if direct_solver:
			inv = a.mat.Inverse(freedofs=freedofs,inverse=inverse)
			gfu.vec.data +=  PreconditionedRichardson(a, f.vec, inv, freedofs )
			
			del inv
		else:
			gfu.vec.data += GMRes(a.mat, f.vec, pre=c.mat, maxsteps=1000, 
									restart=100, tol=1e-10,freedofs=freedofs)

		if condense:
			gfu.vec.data += a.harmonic_extension * gfu.vec
			gfu.vec.data += a.inner_solve * f.vec
	
	CompError()
	Redraw(blocking=True)
	

# ------------------------------ POST-PROCESSING ------------------------------

print("\n")
for key,errs in Errors.items():
	print("{} Errors:".format(key))
	print("----------------------------")
	print("lvl\tError\t\tRate")
	for l in lvls:
		print("{:}\t{:1.3e}\t".format(l,errs[l]), end="")
		if l > 0:
			rate = (log(errs[l-1])-log(errs[l]))/log(2)
			print("{:1.1f}".format(rate))
		else: print("--")
	print("----------------------------\n")


end_time = time.time() - start_time
print("\n----------- Total time: {:02.0f}:{:02.0f}:{:02.0f}:{:02.0f} ----------".format(
    end_time//(24*60*60),end_time%(24*60*60)//(60*60), end_time%3600//60, end_time%60 ) )